/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2001
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
 * mv /usr/lib/cups/filter/CUPSToOmni /usr/lib/cups/filter/CUPSToOmni.orig
 * echo "#!/bin/bash"                              > /usr/lib/cups/filter/CUPSToOmni
 * echo "export OMNI_CUPS_SERVER_PAUSE=1"         >> /usr/lib/cups/filter/CUPSToOmni
 * echo "/usr/lib/cups/filter/CUPSToOmni.orig $*" >> /usr/lib/cups/filter/CUPSToOmni
 * chmod +x /usr/lib/cups/filter/CUPSToOmni
 *
 * (cd CUPS/; make; /bin/cp .libs/CUPSToOmni /usr/lib/cups/filter/CUPSToOmni.orig)
 */

#include "Device.hpp"
#include "Omni.hpp"
#include "JobProperties.hpp"

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <sys/types.h>
#include <fcntl.h>

#include <glib.h>
#include <gmodule.h>

#include <cups/cups.h>
#include <cups/ppd.h>
#include <cups/raster.h>

#define DEBUG_LOGFILE "/tmp/CUPSToOmni.debug.log"

typedef struct _HWMAR {
   float fLeftClip;
   float fBottomClip;
   float fRightClip;
   float fTopClip;
   float fxWidth;
   float fyHeight;
} HWMARGINS;

typedef struct _HWRES {
   float xRes;
   float yRes;
   float fScanDots;  // number of dots in scan line
} HWRESOLUTION;

typedef struct _PRTMODE {
   int iBitCount;
   int iPlanes;
} PRINTMODE;

/*-----------------------------------------------------------------------------*/
/*                                                                             */
/* Function: FindBandSize                                                      */
/*  This function figures out the appropriate band size based on the amount    */
/*  of memory that can be occupied by the band and bitmap information.         */
/*  bitmap, and the specifics of the hardware                                  */
/*                                                                             */
/*-----------------------------------------------------------------------------*/
ULONG
FindBandSize (ULONG ulYHeight,
              ULONG ulXWidth,
              ULONG ulBitsPerPel,
              ULONG ulNumPlanes,
              ULONG ulModulus,
              ULONG ulMemoryLimit)
{
   ULONG ulSizeScanLine;
   ULONG ulMemoryNeeded;
   ULONG ulNumLinesFit;

   if (0 == ulModulus)
   {
      ulModulus = 1;
   }

   // figure out how much memory is needed fore each line
   ulSizeScanLine = ((ulBitsPerPel * ulXWidth + 31) / 32) * ulNumPlanes * 4;

   // Figure out how much memory is needed for the page
   ulMemoryNeeded = ulYHeight * ulSizeScanLine;

   // How many lines can fit in the size given?
   ulNumLinesFit = ulMemoryLimit / ulSizeScanLine;

   if (0 == ulNumLinesFit)
      // Minimum of 1 scan line
      ulNumLinesFit = 1;

   if (ulNumLinesFit <= ulModulus)
      // Not enough lines... Promote it to a modulus.
      ulNumLinesFit = ulModulus;
   else
      // Bump down the number of lines so that it is a modulus.
      ulNumLinesFit -= ulNumLinesFit % ulModulus;

   if ((ulYHeight % ulNumLinesFit) * 100 / ulYHeight <= 15)
   {
      USHORT usBumpUp;

      usBumpUp  = ulYHeight % ulNumLinesFit;
      usBumpUp += ulModulus - 1;
      usBumpUp /= ulModulus;
      usBumpUp *= ulModulus;

      ulNumLinesFit += usBumpUp;
   }

   return ulNumLinesFit;  // return the number of lines we want for the band
}

void
fixEnvironment (FILE *fpCUPSDebug, int fDebugOutput)
{
   static char *pszOmniPATH = "/opt/Omni/bin";
   char        *pszPATH     = 0;

   pszPATH = getenv ("PATH");

#ifdef DEBUG
   if (fDebugOutput)
   {
      fprintf (fpCUPSDebug, "CUPSToOmni: PATH            = %s\n", pszPATH);
      fprintf (fpCUPSDebug, "CUPSToOmni: LD_LIBRARY_PATH = %s\n", getenv ("LD_LIBRARY_PATH"));
   }
#endif

   if (0 == strstr (pszPATH, "Omni/bin"))
   {  //                                   ':'                        '\0'
      int   cbAlloc    = strlen (pszPATH) + 1 + strlen (pszOmniPATH) + 1;
      char *pszNewPATH = 0;

      pszNewPATH = (char *)malloc (cbAlloc);
      if (pszNewPATH)
      {
         int rc = 0;

         sprintf (pszNewPATH, "%s:%s", pszPATH, pszOmniPATH);

         rc = setenv ("PATH", pszNewPATH, 1);

#ifdef DEBUG
         if (fDebugOutput)
         {
            fprintf (fpCUPSDebug, "CUPSToOmni::setenv (\"PATH\", \"%s\") = %d\n", pszNewPATH, rc);
         }
#endif

         free (pszNewPATH);
      }
   }
}

char *
dequoteString (char *pszString)
{
   if (!pszString)
      return 0;

   int   cbNewString = 0;
   char *pszFrom     = pszString;

   while (*pszFrom)
   {
      switch (*pszFrom++)
      {
      case '%': cbNewString++; pszFrom += 2; break;
      default:  cbNewString++; break;
      }
   }

   if (cbNewString)
   {
      cbNewString++;

      char *pszNewString = (char *)calloc (1, cbNewString);
      char *pszTo        = 0;

      if (pszNewString)
      {
         pszFrom = pszString;
         pszTo   = pszNewString;

         while (*pszFrom)
         {
            switch (*pszFrom)
            {
            case '%':
            {
               if (0 == strncmp (pszFrom + 1, "20", 2))
               {
                  *pszTo++ = ' ';
                  pszFrom += 3;
               }
               else if (0 == strncmp (pszFrom + 1, "22", 2))
               {
                  *pszTo++ = '"';
                  pszFrom += 3;
               }
               else if (0 == strncmp (pszFrom + 1, "25", 2))
               {
                  *pszTo++ = '%';
                  pszFrom += 3;
               }
               break;
            }
            default:
            {
               *pszTo++ = *pszFrom++;
               break;
            }
            }
         }

         *pszTo = '\0';

         return pszNewString;
      }
   }

   return 0;
}

int
main (int argc, char *argv[])
{
#define MAX_SCAN_LINES 1024
   PSZRO               pszPPDFilename         = 0; /* PPD environment variable   */
   ppd_file_t         *ppd                    = 0; /* PPD file                   */
   int                 fdInput                = 0; /* File descriptor            */
   cups_raster_t      *pRasterStream          = 0; /* Raster stream to read from */
   cups_page_header_t  pageHeader;                 /* Page header from file      */
   cups_page_header_t  pageHeaderCurrent;          /* Page header from file      */
   char               *pszLibraryName         = 0;
   GModule            *hmodDevice             = 0;
   Device             *pDevice                = 0;
   PFNNEWDEVICEWARGS   pfnNewDeviceWArgs      = 0;
   int                 iPageNumber            = 0;
   unsigned int        uiScanline             = 0;
   unsigned int        iNumScanLines          = MAX_SCAN_LINES;
   PBYTE               pbBits                 = 0;
   PBYTE               pbBitsToSend           = 0;
   PBITMAPINFO2        pbmi2Bitmap            = 0;
   int                 iNumColors             = 0;
   int                 icbBitmapScanLine      = 0;
   int                 iBytesToAllocate       = 0;
   RECTL               rectlPageLocation;
   HWRESOLUTION        hwRes;
   DeviceResolution   *pDevRes                = 0; // Device Resolution stuff
   HardCopyCap        *pHCC                   = 0; // Hard copy information
   unsigned long       ulBytesRead            = 0;
   bool                fFound                 = false;
   int                 i,
                       j,
                       k;
   unsigned int        xRes,
                       yRes;
   PSZRO               pszCommandLine         = 0;
   char               *pszCommandLineArg      = 0;
   bool                fTopOfBand             = false;
   int                 fDebugOutput           = 0;
   int                 fDoOutput              = 0;
   FILE               *fpCUPSDebug            = 0;
   int                 rc                     = 0;
   PSZRO               pszJobProperties       = 0;
   JobProperties       jobProp ("");
   std::ostream&       cerrReal               = std::cerr;

   if (getenv ("OMNI_CUPS_SERVER_PAUSE"))
   {
      // Handy to attach via gdb debugger.  do:
      // (gdb) file omni
      // (gdb) shell ps -efl.  Look for: 100 R lp <pid> 9295 83 85 0 - 1263 - 11:09 ? 00:01:15 bob 44 hamzy tiger.ps 1
      // (gdb) attach <pid>
      // (gdb) set fPause = 0
      // (gdb) c
      bool fPause = true;

      while (fPause)
      {
      }
   }

   // @TBD - figure out how to get error messages into /var/log/cups/error_log
   fprintf (stderr, "INFO: Hello world 1!\n");
   fputs("INFO: Hello world 2!\n", stderr);

#ifdef DEBUG
   fDebugOutput = 1;
#endif

#ifdef DEBUG
   if (fDebugOutput)
   {
      fpCUPSDebug = fopen (DEBUG_LOGFILE, "wb");

      DebugOutput::setErrorStream (fpCUPSDebug);
      DebugOutput::setDebugOutput ("all");
   }
#endif

#ifdef DEBUG
   if (fDebugOutput)
   {
      fprintf (fpCUPSDebug, "CUPSToOmni: fDoOutput = %d\n", fDoOutput);
      for (i = 0; i < argc; i++)
      {
          fprintf (fpCUPSDebug, "CUPSToOmni: Arg %d = \"%s\"\n", i, argv[i]);
      }
   }
#endif

   if (  argc < 5
      || argc > 7
      )
   {
      /*
       * We don't have the correct number of arguments; write an error message
       * and return.
       */
      cerrReal << "ERROR: CUPSToOmni job-id user title copies options [file]" << std::endl;
      cerrReal << "       and argc = " << argc << std::endl;

      rc = 1;
      goto done;
   }

   Omni::initialize ();

   if (6 <= argc)
   {
      pszCommandLine = (PSZRO)argv[5];

      pszCommandLineArg = strstr (pszCommandLine, (PSZRO)"no-output");

      if (pszCommandLineArg)  // enable print output
         fDoOutput = 0;
      else
         fDoOutput = 1;

#ifdef DEBUG
      pszCommandLineArg = strstr (pszCommandLine, (PSZRO)"no-debug");

      if (pszCommandLineArg)  // enable debugging output
         fDebugOutput = 0;
#endif
   }

   fixEnvironment (fpCUPSDebug, fDebugOutput);

   /*
    * Get the PPD file and figure out which driver to use...
    */
   if ((pszPPDFilename = getenv ("PPD")) == NULL)
   {
      cerrReal << "ERROR: Fatal error: PPD environment variable not set!" << std::endl;

      rc = 1;
      goto done;
   }

   if ((ppd = ppdOpenFile (pszPPDFilename)) == NULL)
   {
      cerrReal << "ERROR: Fatal error: Unable to load PPD file " << pszPPDFilename << std::endl;

      rc = 1;
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: successful ppd open\n");
#endif

   if (ppd->shortnickname == NULL)
   {
      cerrReal << "ERROR: Fatal error: No ShortNickName attribute in PPD file " << pszPPDFilename << std::endl;

      rc = 1;
      goto done;
   }

#ifdef DEBUG
   if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: The shortnickname is \"%s\"\n", ppd->shortnickname);
#endif

   pszLibraryName = dequoteString (ppd->shortnickname);

#ifdef DEBUG
   if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: dequoted its \"%s\"\n", pszLibraryName);
#endif

   // Search for the space separator of job properties.  reuse a variable for temp storage.
   pszCommandLineArg = strchr (pszLibraryName, ' ');
   if (pszCommandLineArg)
   {
      // Separate the library name from the job properties
      *pszCommandLineArg = '\0';

      jobProp.setJobProperties (pszCommandLineArg + 1);
   }

   hmodDevice = g_module_open (pszLibraryName, (GModuleFlags)0);

#ifdef DEBUG
   if (fDebugOutput)
   {
      if (hmodDevice)
      {
         fprintf (fpCUPSDebug, "CUPSToOmni: successful driver load\n");
      }
   }
#endif

   if (!hmodDevice)
   {
#ifdef DEBUG
      fprintf (fpCUPSDebug, "CUPSToOmni: Error: unsuccessful driver load\"%s\"\n", g_module_error ());
#endif

      cerrReal << "ERROR: Fatal error: unable to load \""
               << pszLibraryName
               << "\", g_module_error returns \""
               << g_module_error ()
               << "\""
               << std::endl;

      rc = 1;
      goto done;
   }

   g_module_symbol (hmodDevice,
                    "newDeviceW_JopProp_Advanced",
                    (gpointer *)&pfnNewDeviceWArgs);

#ifdef DEBUG
   if (fDebugOutput)
   {
      if (pfnNewDeviceWArgs)
      {
         fprintf (fpCUPSDebug, "CUPSToOmni: set up driver successful\n");
      }
   }
#endif

   if (!pfnNewDeviceWArgs)
   {
#ifdef DEBUG
      fprintf (fpCUPSDebug, "CUPSToOmni: Error: missing newDeviceW_JopProp_Advanced.\n");
#endif

      cerrReal << "ERROR: Fatal error: " << pszLibraryName << " is not a valid library" << std::endl;

      rc = 1;
      goto done;
   }

   /*
    * Open the page stream...
    */
   if (argc == 7)
   {
      if ((fdInput = open (argv[6], O_RDONLY)) == -1)  //open the raster file
      {
         perror ("ERROR: Fatal error: Unable to open raster file - ");
         sleep (1);

         rc = 1;
         goto done;
      }
   }
   else
   {
      fdInput = 0;
   }

   pRasterStream = cupsRasterOpen (fdInput, CUPS_RASTER_READ);

#ifdef DEBUG
   if (fDebugOutput)
   {
      if (!pRasterStream)
      {
         fprintf (fpCUPSDebug, "CUPSToOmni: Warning: cupsRasterOpen Failed.\n");
      }
   }
#endif

#ifdef DEBUG
   if (fDebugOutput) if (pszCommandLine) fprintf (fpCUPSDebug, "CUPSToOmni: command line = \"%s\"\n", pszCommandLine);
#endif

   /*
    * Process pages as needed...
    */
   while (cupsRasterReadHeader (pRasterStream, &pageHeader))
   {
      /***************************************************************************/
      /*                                                                         */
      /* Current Monochrome selection is:                                        */
      /*   -o ColorModel=Gray                                                    */
      /*                                                                         */
      /* We are currently using the FORM_ strings and such to invoke Omni        */
      /*                                                                         */
      /* Formulate a string to pass into the driver based on the following       */
      /* parameter list                                                          */
      /*                                                                         */
      /*   - form:                                                               */
      /*                                                                         */
      /*     form=FORM_LETTER or form=FORM_A4                                    */
      /*                                                                         */
      /*   - resolution:                                                         */
      /*                                                                         */
      /*     resolution=RESOLUTION_360_X_360 or resolution=RESOLUTION_180_X_180  */
      /*                                                                         */
      /*   - media:                                                              */
      /*                                                                         */
      /*     media=MEDIA_PLAIN or media=MEDIA_GLOSSY                             */
      /*                                                                         */
      /*   - dithering:                                                          */
      /*                                                                         */
      /*     dither=DITHER_STUCKI_DIFFUSION or dither=DITHER_STEINBERG_DIFFUSION */
      /*                                                                         */
      /*   - tray:                                                               */
      /*                                                                         */
      /*     tray=TRAY_MANUAL_FEEDER                                             */
      /*                                                                         */
      /*   - printmode:                                                          */
      /*                                                                         */
      /*     printmode=PRINT_MODE_1_ANY (**) or printmode=PRINT_MODE_24_CMYK     */
      /*                                                                         */
      /***************************************************************************/

      // we need to do preliminary driver setup here
      if (iPageNumber == 0)
      {
#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: PageSize = (%d, %d)\n", pageHeader.PageSize[0], pageHeader.PageSize[1]);
#endif

         // get the form name
         fFound = false;
         for (i = 0; i < ppd->num_sizes && !fFound; i++)
         {
            if (  (ppd->sizes[i].width  == pageHeader.PageSize[0])
               && (ppd->sizes[i].length == pageHeader.PageSize[1])
               )
            {
               jobProp.setJobProperty ("form", ppd->sizes[i].name);
               fFound = true;
            }
         }

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: HWResolution = (%d, %d)\n", pageHeader.HWResolution[0], pageHeader.HWResolution[1]);
#endif

         // get the resolution
         fFound = false;
         for (i = 0; i < ppd->num_groups && !fFound; i++)
         {
            for (j = 0; j < ppd->groups[i].num_options && !fFound; j++)
            {
               if (0 == strcmp ("Resolution", ppd->groups[i].options[j].text))
               {
                  for (k = 0; k < ppd->groups[i].options[j].num_choices; k++)
                  {
                     char *pszTemp = 0;

                     // get the string for the [resx resy]setpagedevice
                     pszTemp = strchr (ppd->groups[i].options[j].choices[k].code, '[');

                     if (pszTemp)
                     {
                        pszTemp++;
                        while (!isdigit (*pszTemp))
                        {
                           pszTemp++;
                        }
                        xRes = atoi (pszTemp);
                        while (isdigit (*pszTemp))
                        {
                           pszTemp++;
                        }
                        while (!isdigit (*pszTemp))
                        {
                           pszTemp++;
                        }
                        yRes = atoi (pszTemp);

                        if (  (xRes == pageHeader.HWResolution[0])
                           && (yRes == pageHeader.HWResolution[1])
                           )
                        {
                           jobProp.setJobProperty ("resolution", ppd->groups[i].options[j].choices[k].choice);
                           fFound = true;
                           break;
                        }
                     }
                  }
               }
            }
         }

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: MediaType = %s\n", pageHeader.MediaType);
#endif

         // get the media type
         fFound = !*pageHeader.MediaType; // If there is something to search for then we havent found it yet!
         for (i = 0; i < ppd->num_groups && !fFound; i++)
         {
            for (j = 0; j < ppd->groups[i].num_options && !fFound; j++)
            {
#ifdef DEBUG
///////////////if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: media key = %s\n", ppd->groups[i].options[j].text);
#endif

               if (0 == strcmp ("Media Type", ppd->groups[i].options[j].text))
               {
                  for (k = 0; k < ppd->groups[i].options[j].num_choices; k++)
                  {
#ifdef DEBUG
                     if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: MediaType Val = %s\n", ppd->groups[i].options[j].choices[k].text);
#endif

                     if (0 == strcmp (pageHeader.MediaType, ppd->groups[i].options[j].choices[k].text))
                     {
                        jobProp.setJobProperty ("media", ppd->groups[i].options[j].choices[k].choice);
                        fFound = true;
                        break;
                     }
                  }
               }
            }
         }

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: dither = %s\n", pageHeader.OutputType);
#endif

         // get the dither type
         fFound = !*pageHeader.OutputType; // If there is something to search for then we havent found it yet!
         for (i = 0; i < ppd->num_groups && !fFound; i++)
         {
            for (j = 0; j < ppd->groups[i].num_options && !fFound; j++)
            {
#ifdef DEBUG
///////////////if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: dither key = %s\n", ppd->groups[i].options[j].text);
#endif

               if (0 == strcmp ("Dither", ppd->groups[i].options[j].text))
               {
                  for (k = 0; k < ppd->groups[i].options[j].num_choices; k++)
                  {
#ifdef DEBUG
                     if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: Dither Val = %s\n", ppd->groups[i].options[j].choices[k].text);
#endif

                     if (0 == strcmp (pageHeader.OutputType, ppd->groups[i].options[j].choices[k].text))
                     {
                        jobProp.setJobProperty ("dither", ppd->groups[i].options[j].choices[k].choice);
                        fFound = true;
                        break;
                     }
                  }
               }
            }
         }

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: MediaClass = %s\n", pageHeader.MediaClass);
#endif

         // get the tray selection
         fFound = !*pageHeader.MediaClass; // If there is something to search for then we havent found it yet!
         for (i = 0; i < ppd->num_groups && !fFound; i++)
         {
            for (j = 0; j < ppd->groups[i].num_options && !fFound; j++)
            {
#ifdef DEBUG
///////////////if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: Media Source = %s\n", ppd->groups[i].options[j].text);
#endif

               if (0 == strcmp ("Media Source", ppd->groups[i].options[j].text))
               {
                  for (k = 0; k < ppd->groups[i].options[j].num_choices; k++)
                  {
#ifdef DEBUG
                     if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: Media Source Val = %s\n", ppd->groups[i].options[j].choices[k].text);
#endif

                     if (0 == strcmp (pageHeader.MediaClass, ppd->groups[i].options[j].choices[k].text))
                     {
                        jobProp.setJobProperty ("tray", ppd->groups[i].options[j].choices[k].choice);
                        fFound = true;
                        break;
                     }
                  }
               }
            }
         }

         if (pageHeader.cupsBitsPerPixel == 1)
         {
            jobProp.setJobProperty ("printmode", "PRINT_MODE_1_ANY");
         }
         else
         {
            // @TBD
         }

         pszJobProperties = jobProp.getJobProperties ();

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: pDevice = 0x%08X, Calling NewDeviceWArgs with \"%s\"\n", (int)pDevice, pszJobProperties);
#endif

         // true for non-pdc implementation
         pDevice = pfnNewDeviceWArgs (pszJobProperties, true);

#ifdef DEBUG
         if (fDebugOutput)
         {
            if (pDevice)
            {
               fprintf (fpCUPSDebug, "CUPSToOmni: set up of new driver properties successful\n");
            }
         }
#endif

         if (!pDevice)
         {
#ifdef DEBUG
            fprintf (fpCUPSDebug, "CUPSToOmni: ERROR: Creating a new device with \"%s\" failed\n", pszJobProperties);
#endif

            cerrReal << "ERROR: Fatal error: Creating a new device with \"" << pszJobProperties << "\" failed" << std::endl;

            rc = 1;
            goto done;
         }

         if (pszJobProperties)
         {
            free ((void *)pszJobProperties);
            pszJobProperties = 0;
         }

         pageHeaderCurrent.cupsColorSpace        = pageHeader.cupsColorSpace;
         pageHeaderCurrent.cupsWidth             = pageHeader.cupsWidth;
         pageHeaderCurrent.cupsHeight            = pageHeader.cupsHeight;
         pageHeaderCurrent.cupsBitsPerColor      = pageHeader.cupsBitsPerColor;
         pageHeaderCurrent.cupsBytesPerLine      = pageHeader.cupsBytesPerLine;
         pageHeaderCurrent.cupsBitsPerPixel      = pageHeader.cupsBitsPerPixel;
         strcpy (pageHeaderCurrent.MediaClass, pageHeader.MediaClass);
         strcpy (pageHeaderCurrent.MediaType,  pageHeader.MediaType);
         strcpy (pageHeaderCurrent.OutputType, pageHeader.OutputType);
         pageHeaderCurrent.HWResolution[0]       = pageHeader.HWResolution[0];
         pageHeaderCurrent.HWResolution[1]       = pageHeader.HWResolution[1];
         pageHeaderCurrent.PageSize[0]           = pageHeader.PageSize[0];
         pageHeaderCurrent.PageSize[1]           = pageHeader.PageSize[1];
         pageHeaderCurrent.ImagingBoundingBox[0] = pageHeader.ImagingBoundingBox[0];
         pageHeaderCurrent.ImagingBoundingBox[1] = pageHeader.ImagingBoundingBox[1];
         pageHeaderCurrent.ImagingBoundingBox[2] = pageHeader.ImagingBoundingBox[2];
         pageHeaderCurrent.ImagingBoundingBox[3] = pageHeader.ImagingBoundingBox[3];
      }

#ifdef DEBUG
      if (fDebugOutput)
      {
         fprintf (fpCUPSDebug, "CUPSToOmni: HW Res                 = %u x %u\n", pageHeader.HWResolution[0],
                                                                                 pageHeader.HWResolution[1]);
         fprintf (fpCUPSDebug, "CUPSToOmni: page size              = %d x %d\n", (int)pageHeader.PageSize[0],
                                                                                 (int)pageHeader.PageSize[1]);
         fprintf (fpCUPSDebug, "CUPSToOmni: Bounding Box           = %u %u %u %u\n", pageHeader.ImagingBoundingBox[0],
                                                                                     pageHeader.ImagingBoundingBox[1],
                                                                                     pageHeader.ImagingBoundingBox[2],
                                                                                     pageHeader.ImagingBoundingBox[3]);

         fprintf (fpCUPSDebug, "CUPSToOmni: cupsColorSpace         = %d\n", pageHeader.cupsColorSpace);
         fprintf (fpCUPSDebug, "CUPSToOmni: cupsWidth              = %d\n", pageHeader.cupsWidth);
         fprintf (fpCUPSDebug, "CUPSToOmni: cupsHeight             = %d\n", pageHeader.cupsHeight);
         fprintf (fpCUPSDebug, "CUPSToOmni: cupsBitsPerColor       = %d\n", pageHeader.cupsBitsPerColor);
         fprintf (fpCUPSDebug, "CUPSToOmni: cupsBytesPerLine       = %d\n", pageHeader.cupsBytesPerLine);
         fprintf (fpCUPSDebug, "CUPSToOmni: cupsBitsPerPixel       = %d\n", pageHeader.cupsBitsPerPixel);
         fprintf (fpCUPSDebug, "CUPSToOmni: MediaClass             = %s\n", pageHeader.MediaClass);
         fprintf (fpCUPSDebug, "CUPSToOmni: MediaType              = %s\n", pageHeader.MediaType);
         fprintf (fpCUPSDebug, "CUPSToOmni: OutputType             = %s\n", pageHeader.OutputType);
      }
#endif

//////iNumScanLines = omni::min (MAX_SCAN_LINES, (int)(pageHeader.cupsHeight - uiScanline));
      iNumColors    = 1 << pageHeaderCurrent.cupsBitsPerPixel;

#ifdef DEBUG
      if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: Number of colors       = %d\n", iNumColors);
#endif

      if (!pbmi2Bitmap)
      {
         // Allocate a full bitmap info 2 and bitmap header message
         iBytesToAllocate = sizeof (BITMAPINFO2);
         if (256 >= iNumColors)
            iBytesToAllocate += (iNumColors - 1) * sizeof (RGB2);

         pbmi2Bitmap = (PBITMAPINFO2)calloc (1, iBytesToAllocate);
         if (!pbmi2Bitmap)
         {
#ifdef DEBUG
            fprintf (fpCUPSDebug, "CUPSToOmni: Error: failed to allocate bitmap info memory!\n");
#endif

            cerrReal << "ERROR: Fatal error: failed to allocate bitmap info memory!" << std::endl;

            rc = 1;
            goto done;
         }

         // Initialize it
         pbmi2Bitmap->cbFix         = sizeof (BITMAPINFO2)
                                    - sizeof (((PBITMAPINFO2)0)->argbColor);

         pbmi2Bitmap->cx            = pageHeaderCurrent.cupsWidth;
         pbmi2Bitmap->cPlanes       = 1;
         pbmi2Bitmap->cBitCount     = pageHeaderCurrent.cupsBitsPerPixel;
         pbmi2Bitmap->cclrUsed      = 1 << pbmi2Bitmap->cBitCount;
         pbmi2Bitmap->cclrImportant = 0;                            // All are important

         // Read the color table, if there is one
         // @TBD
      }

      icbBitmapScanLine = (((pbmi2Bitmap->cx) * pbmi2Bitmap->cBitCount + 31) >> 5) << 2;

      iPageNumber++;
      uiScanline = 0;

      if (1 == iPageNumber)
      {
         pDevice->beginJob ();
      }

      // get the hardware resolution
      pDevRes = pDevice->getCurrentResolution ();

      hwRes.xRes      = pDevRes->getXRes ();
      hwRes.yRes      = pDevRes->getYRes ();
      hwRes.fScanDots = pDevRes->getScanlineMultiple (); // need head pass information

      // get the hardware page margins information
      // **** might not currently need this info
      pHCC = pDevice->getCurrentForm ()->getHardCopyCap ();

      if (iPageNumber == 1)
      {
         // need to calculate the correct bandsize needed for the device
         iNumScanLines = FindBandSize (pageHeaderCurrent.cupsHeight,
                                       pageHeaderCurrent.cupsWidth,
                                       24,           // always use 24 bits for source bitmap (for now)
                                       1,            // number of planes
                                       (unsigned)hwRes.fScanDots,
                                       4000 * 1024); // eight meg buffer
      }

#ifdef DEBUG
      if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: bandsize               = %d\n", iNumScanLines);
#endif

      if (!pbBits)
      {
         iBytesToAllocate = icbBitmapScanLine * iNumScanLines;

         pbBits = (PBYTE)calloc (1, iBytesToAllocate);
         if (!pbBits)
         {
#ifdef DEBUG
            fprintf (fpCUPSDebug, "CUPSToOmni: Error: failed to allocate bitmap memory!\n");
#endif

            cerrReal << "ERROR: Fatal error: failed to allocate bitmap memory!" << std::endl;

            rc = 1;
            goto done;
         }
      }

#ifdef DEBUG
      if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: cupsBytesPerLine = %d, icbBitmapScanLine = %d\n", pageHeaderCurrent.cupsBytesPerLine, icbBitmapScanLine);
#endif

      uiScanline = pageHeaderCurrent.cupsHeight - 1;

      rectlPageLocation.xLeft  = 0;
      rectlPageLocation.xRight = pbmi2Bitmap->cx;

      while (uiScanline >= 0)
      {
#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: uiScanline          = %u\n", uiScanline);
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: pbmi2Bitmap->cx     = %d\n", pbmi2Bitmap->cx);
#endif

         pbmi2Bitmap->cy = (iNumScanLines < uiScanline) ? iNumScanLines : uiScanline;

         rectlPageLocation.yTop    = uiScanline;
         rectlPageLocation.yBottom = uiScanline - pbmi2Bitmap->cy;

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: pbmi2Bitmap->cy     = %d\n", pbmi2Bitmap->cy);
#endif

         ulBytesRead = cupsRasterReadPixels (pRasterStream,
                                             pbBits,
                                             pageHeaderCurrent.cupsBytesPerLine * pbmi2Bitmap->cy);

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: bytes to read       = %ld\n", (long)(pageHeaderCurrent.cupsBytesPerLine * pbmi2Bitmap->cy));
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: bytes read          = %ld\n", ulBytesRead);
#endif

         if (ulBytesRead != (ULONG)(pageHeaderCurrent.cupsBytesPerLine * pbmi2Bitmap->cy))
         {
#ifdef DEBUG
            fprintf (fpCUPSDebug, "CUPSToOmni: Error: reading bitmap data\n");
            fprintf (fpCUPSDebug, "CUPSToOmni: read %ld bytes of raster data\n", (long)ulBytesRead);
            fprintf (fpCUPSDebug, "CUPSToOmni: needed %ld bytes of raster data\n", (long)(pageHeaderCurrent.cupsBytesPerLine * pbmi2Bitmap->cy));
#endif

            cerrReal << "Error reading bitmap data" << std::endl;
            cerrReal << "read " << ulBytesRead <<  " bytes of raster data" << std::endl;
            cerrReal << "needed " << (pageHeaderCurrent.cupsBytesPerLine * pbmi2Bitmap->cy) << " bytes of raster data\n" << std::endl;

            rc = 1;
            break;
         }

         // Test to see if CUPS had anymore data for us even though the page routine is called.
#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: ************** Page Number = %d\n", iPageNumber);
#endif

         if (  (iPageNumber != 1)
            && ulBytesRead
            && fTopOfBand
            )
         {
            fTopOfBand = false;

            pDevice->newFrame ();

#ifdef DEBUG
            if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: doing new frame\n");
#endif
         }

#ifdef DEBUG
         fflush (fpCUPSDebug);
#endif

         pbBitsToSend = (PBYTE)malloc (icbBitmapScanLine * pbmi2Bitmap->cy);

         if (!pbBitsToSend)
         {
#ifdef DEBUG
            fprintf (fpCUPSDebug, "CUPSToOmni: Error: failed to allocate flip bitmap memory!\n");
#endif

            cerrReal << "Error: failed to allocate flip bitmap memory!" << std::endl;

            rc = 1;
            goto done;
         }

         if (pageHeaderCurrent.cupsBitsPerPixel != 1)
         {
            // flip and rotate BGR to RGB in the output band
            for (i = 0; i < pbmi2Bitmap->cy; i++)
            {
               for (j = 0; j < (int)pageHeaderCurrent.cupsBytesPerLine - 2; j += 3)
               {
                  int p = j + 2;

                  pbBitsToSend[i*icbBitmapScanLine+j] =    // b --> r
                                  pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)+p];
                  pbBitsToSend[i*icbBitmapScanLine+p] =    // r --> b
                                  pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)+j];
                  pbBitsToSend[i*icbBitmapScanLine+j+1] =  // g --> g
                                  pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)+j+1];
               }
            }

#ifdef DEBUG
            if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: middle cy = %u\n", pbmi2Bitmap->cy);
#endif
         }
         else
         {
            for (i = 0; i < pbmi2Bitmap->cy; i++)
            {
               memcpy ((void *)&pbBitsToSend[i*icbBitmapScanLine],
                       (void *)&pbBits[pageHeaderCurrent.cupsBytesPerLine * (pbmi2Bitmap->cy-1 -i)],
                       pageHeaderCurrent.cupsBytesPerLine);
            }
         }

         uiScanline -= pbmi2Bitmap->cy;  // add in number of lines read

         if (fDoOutput)
         {
             pDevice->rasterize (pbBitsToSend,
                                 pbmi2Bitmap,
                                 &rectlPageLocation,
                                 BITBLT_BITMAP);
         }

         if (pbBitsToSend)
         {
            free (pbBitsToSend);
            pbBitsToSend = 0;
         }

#ifdef DEBUG
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: End pbmi2Bitmap->cy = %d\n", pbmi2Bitmap->cy);
         if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: End uiScanline      = %u\n", uiScanline);
#endif

         if (!uiScanline)
            break;

      }  // end while to read raster data

      fTopOfBand = true;

#ifdef DEBUG
      if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: Through first loop\n");
#endif
   }  // end read of ppd file

done:

#ifdef DEBUG
   if (fDebugOutput) fprintf (fpCUPSDebug, "CUPSToOmni: dropping out\n");
#endif

   if (  pDevice
      && 1 <= iPageNumber
      )
   {
      pDevice->endJob ();
   }

   /*
    * Close the raster stream...
    */
   if (pRasterStream)
   {
      cupsRasterClose (pRasterStream);
      pRasterStream = 0;
   }

   if (fdInput != 0)
   {
      close (fdInput);
      fdInput = 0;
   }

   if (ppd)
   {
      ppdClose (ppd);
      ppd = 0;
   }

   if (pszLibraryName)
   {
      free (pszLibraryName);
      pszLibraryName = 0;
   }
   if (pbmi2Bitmap)
   {
      free (pbmi2Bitmap);
      pbmi2Bitmap = 0;
   }
   if (pbBitsToSend)
   {
      free (pbBitsToSend);
      pbBitsToSend = 0;
   }
   if (pbBits)
   {
      free (pbBits);
      pbBits = 0;
   }
   if (pszJobProperties)
   {
      free ((void *)pszJobProperties);
      pszJobProperties = 0;
   }

   delete pDevice; pDevice = 0;

   if (hmodDevice)
   {
      g_module_close (hmodDevice);
      hmodDevice = 0;
   }

#ifdef DEBUG
   if (fDebugOutput)
   {
      fprintf (fpCUPSDebug, "CUPSToOmni: Closing output stream\n");

      if (fpCUPSDebug)
      {
         fclose (fpCUPSDebug);

         chmod (DEBUG_LOGFILE, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
      }
   }
#endif

   if (0 == iPageNumber)
   {
      cerrReal << "ERROR: CUPSToOmni No pages printed!" << std::endl;
   }
   else
   {
      cerrReal << "INFO: CUPSToOmni Ready to print." << std::endl;
   }

   Omni::terminate ();

   return rc;
}
