/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "IBM_5577_Instance.hpp"
#include "defines.hpp"

#include <iostream>
#include <sstream>

const static bool fTestNoCompression = true;

DeviceInstance *
createInstance (PrintDevice *pDevice)
{
   return new IBM_5577_Instance (pDevice);
}

void
deleteInstance (DeviceInstance *pInstance)
{
   delete pInstance;
}

IBM_5577_Instance::
IBM_5577_Instance (PrintDevice *pDevice)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::IBM_5577_Instance ()" << endl;
#endif

   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;
}

IBM_5577_Instance::
~IBM_5577_Instance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::~IBM_5577_Instance ()" << endl;
#endif
}

void IBM_5577_Instance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;

   fRET_d         = TSTATE_NONE;
   fEconoMode_d   = TSTATE_NONE;
   fPageProtect_d = TSTATE_NONE;
   fJamRecovery_d = TSTATE_NONE;
   fStaple_d      = TSTATE_NONE;
   fOffset_d      = TSTATE_NONE;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::initializeInstance ()" << endl;
#endif
}

static char *apszDeviceJobPropertyKeys[] = {
   "RET",
   "EconoMode",
   "PageProtect",
   "JamRecovery",
   "Staple",
   "Offset"
};
#define JOBPROP_RET              apszDeviceJobPropertyKeys[0]
#define JOBPROP_ECONOMODE        apszDeviceJobPropertyKeys[1]
#define JOBPROP_PAGEPROTECT      apszDeviceJobPropertyKeys[2]
#define JOBPROP_JAMRECOVERY      apszDeviceJobPropertyKeys[3]
#define JOBPROP_STAPLE           apszDeviceJobPropertyKeys[4]
#define JOBPROP_OFFSET           apszDeviceJobPropertyKeys[5]

class DeviceJobPropertyKeyEnumerator : public Enumeration
{
public:
   DeviceJobPropertyKeyEnumerator (char **apszKeys, int iDimOfKeys)
   {
      iWhichKey_d = iDimOfKeys;
      apszKeys_d  = apszKeys;
   }

   virtual bool hasMoreElements ()
   {
      return iWhichKey_d > 0;
   }

   virtual void * nextElement ()
   {
      if (0 < iWhichKey_d)
      {
         iWhichKey_d--;
         return apszKeys_d[iWhichKey_d];
      }

      return 0;
   }

private:
   int    iWhichKey_d;
   char **apszKeys_d;
};

Enumeration * IBM_5577_Instance::
listDeviceJobPropertyKeys ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "IBM_5577_Instance::" << __FUNCTION__ << " ()" << endl;
#endif

   return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeys, dimof (apszDeviceJobPropertyKeys));
}

class KeyEnumerator : public Enumeration
{
public:
   KeyEnumerator (const char *pszKeys)
   {
      pszKeys_d    = 0;
      pszCurrent_d = 0;

      if (  pszKeys
         && *pszKeys
         )
      {
         pszKeys_d = (char *)malloc (strlen (pszKeys) + 1);
         if (pszKeys_d)
         {
            strcpy (pszKeys_d, pszKeys);
            pszCurrent_d = pszKeys_d;
         }
      }
   }

   virtual ~KeyEnumerator ()
   {
      if (pszKeys_d)
      {
         free (pszKeys_d);
         pszKeys_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return pszCurrent_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = pszCurrent_d;

      if (  pszCurrent_d
         && *pszCurrent_d
         )
      {
         char *pszSpace = strchr (pszCurrent_d, ' ');

         if (pszSpace)
         {
            *pszSpace = '\0';
            pszCurrent_d = pszSpace + 1;
            while (' ' == *pszCurrent_d)
            {
               pszCurrent_d++;
            }
            if (!*pszCurrent_d)
            {
               pszCurrent_d = 0;
            }
         }
         else
         {
            pszCurrent_d = 0;
         }
      }
      else
      {
         pszCurrent_d = 0;
      }

      return pvRet;
   }

private:
   char *pszKeys_d;
   char *pszCurrent_d;
};

Enumeration * IBM_5577_Instance::
listKeyValues (char *pszKey)
{
   ostringstream oss;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "IBM_5577_Instance::" << __FUNCTION__ << " (" << (pszKey ? pszKey : "") << ")" << endl;
#endif

   if (  0 == strcasecmp (pszKey, JOBPROP_RET)
      || 0 == strcasecmp (pszKey, JOBPROP_ECONOMODE)
      || 0 == strcasecmp (pszKey, JOBPROP_PAGEPROTECT)
      || 0 == strcasecmp (pszKey, JOBPROP_JAMRECOVERY)
      || 0 == strcasecmp (pszKey, JOBPROP_STAPLE)
      || 0 == strcasecmp (pszKey, JOBPROP_OFFSET)
      )
   {
      oss << "on off none";
   }

   return new KeyEnumerator (oss.str ().c_str ());
}

bool IBM_5577_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::handleDeviceJobProperty ("
                          << (pszKey ? pszKey : "") << ", "
                          << (pszValue ? pszValue : "") << ")"
                          << endl;
#endif

   if (0 == strcasecmp (pszKey, JOBPROP_RET))
   {
      if (0 == strcmp (pszValue, "on"))
      {
         fRET_d = TSTATE_ON;
         return true;
      }
      else if (0 == strcmp (pszValue, "off"))
      {
         fRET_d = TSTATE_OFF;
         return true;
      }
      else if (0 == strcmp (pszValue, "none"))
      {
         fRET_d = TSTATE_NONE;
         return true;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_ECONOMODE))
   {
      if (0 == strcmp (pszValue, "on"))
      {
         fEconoMode_d = TSTATE_ON;
         return true;
      }
      else if (0 == strcmp (pszValue, "off"))
      {
         fEconoMode_d = TSTATE_OFF;
         return true;
      }
      else if (0 == strcmp (pszValue, "none"))
      {
         fEconoMode_d = TSTATE_NONE;
         return true;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_PAGEPROTECT))
   {
      if (0 == strcmp (pszValue, "on"))
      {
         fPageProtect_d = TSTATE_ON;
         return true;
      }
      else if (0 == strcmp (pszValue, "off"))
      {
         fPageProtect_d = TSTATE_OFF;
         return true;
      }
      else if (0 == strcmp (pszValue, "none"))
      {
         fPageProtect_d = TSTATE_NONE;
         return true;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_JAMRECOVERY))
   {
      if (0 == strcmp (pszValue, "on"))
      {
         fJamRecovery_d = TSTATE_ON;
         return true;
      }
      else if (0 == strcmp (pszValue, "off"))
      {
         fJamRecovery_d = TSTATE_OFF;
         return true;
      }
      else if (0 == strcmp (pszValue, "none"))
      {
         fJamRecovery_d = TSTATE_NONE;
         return true;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_STAPLE))
   {
      if (0 == strcmp (pszValue, "on"))
      {
         fStaple_d = TSTATE_ON;
         return true;
      }
      else if (0 == strcmp (pszValue, "off"))
      {
         fStaple_d = TSTATE_OFF;
         return true;
      }
      else if (0 == strcmp (pszValue, "none"))
      {
         fStaple_d = TSTATE_NONE;
         return true;
      }
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_OFFSET))
   {
      if (0 == strcmp (pszValue, "on"))
      {
         fOffset_d = TSTATE_ON;
         return true;
      }
      else if (0 == strcmp (pszValue, "off"))
      {
         fOffset_d = TSTATE_OFF;
         return true;
      }
      else if (0 == strcmp (pszValue, "none"))
      {
         fOffset_d = TSTATE_NONE;
         return true;
      }
   }

   return false;
}

string * IBM_5577_Instance::
getJobPropertyType (char *pszKey)
{
   TRISTATE state = TSTATE_NONE;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "IBM_5577_Instance::" << __FUNCTION__ << " (" << (pszKey ? pszKey : "") << ")" << endl;
#endif

   if (0 == strcasecmp (pszKey, JOBPROP_RET))
   {
      state = fRET_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_ECONOMODE))
   {
      state = fEconoMode_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_PAGEPROTECT))
   {
      state = fPageProtect_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_JAMRECOVERY))
   {
      state = fJamRecovery_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_STAPLE))
   {
      state = fStaple_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_OFFSET))
   {
      state = fOffset_d;
   }
   else
   {
      return 0;
   }

   switch (fOffset_d)
   {
   case TSTATE_NONE: return new string ("string none");
   case TSTATE_ON:   return new string ("string on");
   case TSTATE_OFF:  return new string ("string off");
   default:          return 0;
   }

   return 0;
}

string * IBM_5577_Instance::
getJobProperty (char *pszKey)
{
   TRISTATE state = TSTATE_NONE;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "IBM_5577_Instance::" << __FUNCTION__ << " (" << (pszKey ? pszKey : "") << ")" << endl;
#endif

   if (0 == strcasecmp (pszKey, JOBPROP_RET))
   {
      state = fRET_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_ECONOMODE))
   {
      state = fEconoMode_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_PAGEPROTECT))
   {
      state = fPageProtect_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_JAMRECOVERY))
   {
      state = fJamRecovery_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_STAPLE))
   {
      state = fStaple_d;
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_OFFSET))
   {
      state = fOffset_d;
   }
   else
   {
      return 0;
   }

   switch (fOffset_d)
   {
   case TSTATE_NONE: return new string ("none");
   case TSTATE_ON:   return new string ("on");
   case TSTATE_OFF:  return new string ("off");
   default:          return 0;
   }

   return 0;
}

string * IBM_5577_Instance::
translateKeyValue (char *pszKey,
                   char *pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) DebugOutput::getErrorStream () << "IBM_5577_Instance::" << __FUNCTION__ << " ("
                          << (pszKey ? pszKey : "") << ", "
                          << (pszValue ? pszValue : "") << ")"
                          << endl;
#endif

   char   *pszRetKey = 0;
   string *pRet      = 0;

   if (0 == strcasecmp (pszKey, JOBPROP_RET))
   {
      pszRetKey = StringResource::getString (pDevice_d->getLanguageResource (),
                                             StringResource::STRINGGROUP_DEVICE_COMMON,
                                             StringResource::DEVICE_COMMON_RET);
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_ECONOMODE))
   {
      pszRetKey = StringResource::getString (pDevice_d->getLanguageResource (),
                                             StringResource::STRINGGROUP_DEVICE_COMMON,
                                             StringResource::DEVICE_COMMON_ECONO_MODE);
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_PAGEPROTECT))
   {
      pszRetKey = StringResource::getString (pDevice_d->getLanguageResource (),
                                             StringResource::STRINGGROUP_DEVICE_COMMON,
                                             StringResource::DEVICE_COMMON_PAGE_PROTECT);
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_JAMRECOVERY))
   {
      pszRetKey = StringResource::getString (pDevice_d->getLanguageResource (),
                                             StringResource::STRINGGROUP_DEVICE_COMMON,
                                             StringResource::DEVICE_COMMON_JAM_RECOVERY);
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_STAPLE))
   {
      pszRetKey = StringResource::getString (pDevice_d->getLanguageResource (),
                                             StringResource::STRINGGROUP_DEVICE_COMMON,
                                             StringResource::DEVICE_COMMON_STAPLE);
   }
   else if (0 == strcasecmp (pszKey, JOBPROP_OFFSET))
   {
      pszRetKey = StringResource::getString (pDevice_d->getLanguageResource (),
                                             StringResource::STRINGGROUP_DEVICE_COMMON,
                                             StringResource::DEVICE_COMMON_OFFSET);
   }

   if (pszRetKey)
   {
      pRet = new string (pszRetKey);
   }

   if (  pszValue
      && pRet
      )
   {
      *pRet += "=";
      *pRet += pszValue;
   }

   return pRet;
}

bool IBM_5577_Instance::
hasDeviceOption (char *pszDeviceOption)
{
   if (  0 == strcmp (pszDeviceOption, "SUPPORTS_PJL")
      || 0 == strcmp (pszDeviceOption, "SUPPORTS_EMULATIONMODE")
      )
   {
      return true;
   }

   return false;
}

void IBM_5577_Instance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::setupPrinter ()" << endl;
#endif

   fHaveSetupPrinter_d = true;

   DeviceResolution *pRes      = getCurrentResolution ();
   DeviceForm       *pDF       = getCurrentForm ();
   HardCopyCap      *pHCC      = pDF->getHardCopyCap ();
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;
}

bool IBM_5577_Instance::
beginJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;
   DeviceForm       *pDF       = getCurrentForm ();
   DeviceTray       *pDT       = getCurrentTray ();

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::beginJob ()" << endl;
#endif

   if (hasDeviceOption ("SUPPORTS_PJL"))
   {
      pCmd = pCommands->getCommandData ("cmdPJLSignature");
      if (pCmd)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPJLSignature = " << *pCmd << endl;
#endif

         pDevice_d->sendBinaryDataToDevice (pCmd);
      }

      pCmd = pCommands->getCommandData ("cmdBeginJob");
      if (pCmd)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputInstance ()) cerr << "cmdBeginJob = " << *pCmd << endl;
#endif

         pDevice_d->sendBinaryDataToDevice (pCmd);
      }

      // @TBD Job prop stuff

      pCmd = pCommands->getCommandData ("cmdEnterLanguage");
      if (pCmd)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputInstance ()) cerr << "cmdEnterLanguage = " << *pCmd << endl;
#endif

         pDevice_d->sendBinaryDataToDevice (pCmd);
      }

      // @TBD Job prop stuff
   }
   else if (hasDeviceOption ("SUPPORTS_EMULATIONMODE"))
   {
      pCmd = pCommands->getCommandData ("cmd5577Mode");
      if (pCmd)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputInstance ()) cerr << "cmd5577Mode = " << *pCmd << endl;
#endif

         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
   }

   pCmd = pCommands->getCommandData ("cmdInit");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdInit = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   // @TBD hardware copies

   DeviceOrientation *pDO = getCurrentOrientation ();

   if (pDO->isID (DeviceOrientation::ORIENTATION_PORTRAIT))
   {
      pCmd = pCommands->getCommandData ("cmdPortrait");
      if (pCmd)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPortrait = " << *pCmd << endl;
#endif

         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
   }
   else if (pDO->isID (DeviceOrientation::ORIENTATION_LANDSCAPE))
   {
      pCmd = pCommands->getCommandData ("cmdLandscape");
      if (pCmd)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputInstance ()) cerr << "cmdLandscape = " << *pCmd << endl;
#endif

         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
   }

   /* form select */
   pCmd = pDF->getData ();
   pDevice_d->sendBinaryDataToDevice (pCmd);

#ifndef RETAIL
 if (DebugOutput::shouldOutputInstance ()) cerr << "pDT->getID(0)   = " << pDT->getID() << endl;
 if (DebugOutput::shouldOutputInstance ()) cerr << "pDF->getID(0)   = " << pDF->getID() << endl;
#endif
   /* tray select */
   if (pDT->getID () == DeviceTray::TRAY_AUTO ||
       pDT->getID () == DeviceTray::TRAY_FRONT_CONTINUOUS ) // for 5577 continuous mode
   {
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::beginJob2 ()" << endl;

      unsigned char tray;
      unsigned int  length = 0;
      unsigned char n1, n2;                                 // for 5577 form length set

#ifndef RETAIL
 if (DebugOutput::shouldOutputInstance ()) cerr << "pDF->getID(1)   = " << pDF->getID() << endl;
#endif
      switch (pDF->getID ())
      {
         case DeviceForm::FORM_A3:           tray = 0x02;
                                             break;
         case DeviceForm::FORM_JIS_B4:       tray = 0x02;
                                             break;
         case DeviceForm::FORM_A4:           tray = 0x02;
                                             break;
         case DeviceForm::FORM_JIS_B5:       tray = 0x02;
                                             break;
         case DeviceForm::FORM_A5:           tray = 0x02;
                                             break;
         case DeviceForm::FORM_LETTER:       tray = 0x02;
                                             break;
         case DeviceForm::FORM_LEGAL:        tray = 0x02;
                                             break;
         case DeviceForm::FORM_HAGAKI_CARD:  tray = 0x02;
                                             break;
         case DeviceForm::FORM_10_X_11:
         case DeviceForm::FORM_11_X_12:
         case DeviceForm::FORM_11_X_15:
         case DeviceForm::FORM_12_X_19:
         case DeviceForm::FORM_15_X_11:      // used mainly in Japan
         case DeviceForm::FORM_5_X_7:        tray = 0x01;

           length = (unsigned int)( (double)pDF->getCy() / 25400.0 * 6.0);
	   n1 = ( length & 0xff00) >> 8;
	   n2 = ( length & 0x00ff);
#ifndef RETAIL
	   if (DebugOutput::shouldOutputInstance ()) cerr << "size Cx = " << pDF->getCx() << endl;
	   if (DebugOutput::shouldOutputInstance ()) cerr << "size Cy = " << pDF->getCy() << endl;
	   if (DebugOutput::shouldOutputInstance ()) cerr << "length  = " << length << endl;
	   // printf("size 0xn1n2 = 0x%x%x\n", n1, n2);
#endif
                                             break;
         default:                            tray = 0x01;
                                             break;
      }

      pCmd = pCommands->getCommandData ("cmdRenTanChg");
      pDevice_d->sendPrintfToDevice (pCmd, tray);

      if ( length ) {
	pCmd = pCommands->getCommandData ("cmdFormLength");
	pDevice_d->sendPrintfToDevice (pCmd, n1, n2);
      }

   }
   else
   {
      pCmd = pDT->getData ();
      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool IBM_5577_Instance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::beginJob (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool IBM_5577_Instance::
newFrame ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::newFrame ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool IBM_5577_Instance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::newFrame (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The orientation is = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool IBM_5577_Instance::
endJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::endJob ()" << endl;
#endif

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   pCmd = pCommands->getCommandData ("cmdTerm");
   if (pCmd)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdTerm = " << *pCmd << endl;
#endif

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool IBM_5577_Instance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputInstance ()) cerr << "IBM_5577_Instance::endJob ()" << endl;
#endif

   // @TBD
   return false;
}

string IBM_5577_Instance::
toString (ostringstream& oss)
{
   oss << "{ "
       << DeviceInstance::toString (oss) // Add parent's output as well
       << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const IBM_5577_Instance& const_self)
{
   IBM_5577_Instance& self = const_cast<IBM_5577_Instance&>(const_self);
   ostringstream       oss;

   os << self.toString (oss);

   return os;
}
