/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <XMLDeviceResolution.hpp>

XMLDeviceResolution::
XMLDeviceResolution (Device      *pDevice,
                     int          id,
                     int          iXRes,
                     int          iYRes,
                     int          iXInternalRes,
                     int          iYInternalRes,
                     BinaryData  *pbdData,
                     int          iCapabilities,
                     int          iDestinationBitsPerPel,
                     int          iScanlineMultiple,
                     xmlNodePtr   node)
   : DeviceResolution (pDevice,
                       id,
                       iXRes,
                       iYRes,
                       iXInternalRes,
                       iYInternalRes,
                       pbdData,
                       iCapabilities,
                       iDestinationBitsPerPel,
                       iScanlineMultiple)
{
   node_d        = node;
   pszDeviceID_d = 0;
}

XMLDeviceResolution::
~XMLDeviceResolution ()
{
   if (pszDeviceID_d)
   {
      free ((void *)pszDeviceID_d);
      pszDeviceID_d = 0;
   }
}

DeviceResolution * XMLDeviceResolution::
createV (int id)
{
   return XMLDeviceResolution::create (pDevice_d, id);
}

DeviceResolution * XMLDeviceResolution::
create (Device *pDevice,
        PSZCRO  pszId)
{
   int id = DeviceResolution::nameToID (pszId);

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceResolution ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": id = " << id << std::endl;
#endif

   if (DeviceResolution::RESOLUTION_UNLISTED != id)
   {
      return create (pDevice, id);
   }

   return 0;
}

DeviceResolution * XMLDeviceResolution::
create (Device *pDevice, int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

   if (!pXMLDevice)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceResolution ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": !pXMLDevice " << std::endl;
#endif

      return 0;
   }

   xmlDocPtr         docDeviceResolutions  = pXMLDevice->getDocResolutions ();
   xmlNodePtr        rootDeviceResolutions = xmlDocGetRootElement (docDeviceResolutions);
   xmlNodePtr        elmDeviceResolution   = 0;
   DeviceResolution *pDRRet                = 0;

   if (!rootDeviceResolutions)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceResolution ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": !rootDeviceResolutions " << std::endl;
#endif

      return 0;
   }

   elmDeviceResolution = firstNode (rootDeviceResolutions);
   if (!elmDeviceResolution)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceResolution ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": !elmDeviceResolution " << std::endl;
#endif

      return 0;
   }

   elmDeviceResolution = firstNode (elmDeviceResolution->children);

   while (  elmDeviceResolution
         && !pDRRet
         )
   {
      PSZCRO pszResolution = getXMLContentString (elmDeviceResolution,
                                                  docDeviceResolutions,
                                                  "name");

      if (pszResolution)
      {
         int idResolution = DeviceResolution::nameToID (pszResolution);

         if (idResolution == id)
         {
            int         iXRes                  = 0;
            int         iYRes                  = 0;
            int         iXInternalRes          = 0;
            int         iYInternalRes          = 0;
            PSZRO       pszCommand             = 0;
            BinaryData *pbdData                = 0;
            PSZRO       pszCapabilities        = 0;
            int         iCapabilities          = 0;
            int         iDestinationBitsPerPel = 0;
            int         iScanlineMultiple      = 0;

            try
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceResolution ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": Creating \"" << pszResolution << "\"" << std::endl;
#endif

               // Read in the X resolution
               iXRes = getXMLContentInt (elmDeviceResolution, docDeviceResolutions, "xRes");
               // Read in the Y resolution
               iYRes = getXMLContentInt (elmDeviceResolution, docDeviceResolutions, "yRes");
               // Read in the internal X resolution
               iXInternalRes = getXMLContentInt (elmDeviceResolution, docDeviceResolutions, "xInternalRes", false);
               // Read in the internal X resolution
               iYInternalRes = getXMLContentInt (elmDeviceResolution, docDeviceResolutions, "yInternalRes", false);

               // Read in the command
               pszCommand = getXMLContentString (elmDeviceResolution, docDeviceResolutions, "command");

               if (pszCommand)
               {
                  byte       *pbData          = 0;
                  int         cbData          = 0;

                  if (XMLDevice::parseBinaryData (pszCommand,
                                                  &pbData,
                                                  &cbData))
                  {
                     pbdData = new BinaryData (pbData, cbData);
                  }

                  free ((void *)pszCommand);
               }

               // Read in the capability
               pszCapabilities = getXMLContentString (elmDeviceResolution, docDeviceResolutions, "resolutionCapability");

               if (pszCapabilities)
               {
                  iCapabilities = DeviceResolution::getReservedValue (pszCapabilities);

                  free ((void *)pszCapabilities);
               }

               // Read in the DestinationBitsPerPel
               iScanlineMultiple = getXMLContentInt (elmDeviceResolution, docDeviceResolutions, "resolutionDestinationBitsPerPel");
               // Read in the ScanlineMultiple
               iScanlineMultiple = getXMLContentInt (elmDeviceResolution, docDeviceResolutions, "resolutionScanlineMultiple");

#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceResolution ())
               {
                  DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": iXRes = " << iXRes << std::endl;
                  DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": iYRes = " << iYRes << std::endl;
                  DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": iXInternalRes = " << iXInternalRes << std::endl;
                  DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": iYInternalRes = " << iYInternalRes << std::endl;
                  if (pbdData)
                     DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": pbdData = " << *pbdData << std::endl;
                  else
                     DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": pbdData is null!" << std::endl;
                  DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": iCapabilities = " << iCapabilities << std::endl;
                  DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": iDestinationBitsPerPel = " << iDestinationBitsPerPel << std::endl;
                  DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": iScanlineMultiple = " << iScanlineMultiple << std::endl;
               }
#endif

               // Create the object
               pDRRet = new XMLDeviceResolution (pDevice,
                                                 id,
                                                 iXRes,
                                                 iYRes,
                                                 iXInternalRes,
                                                 iYInternalRes,
                                                 pbdData,
                                                 iCapabilities,
                                                 iDestinationBitsPerPel,
                                                 iScanlineMultiple,
                                                 elmDeviceResolution);
            }
            catch (std::string *pstringError)
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputXMLDeviceResolution ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": Error: " << *pstringError << std::endl;
#endif

               delete pstringError;
            }
         }

         free ((void *)pszResolution);
      }

      elmDeviceResolution = nextNode (elmDeviceResolution);
   }


#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDeviceResolution ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": returning " << pDRRet << std::endl;
#endif

   return pDRRet;
}

PSZCRO XMLDeviceResolution::
getDeviceID ()
{
   if (!pszDeviceID_d)
   {
      if (node_d)
      {
         pszDeviceID_d = getXMLContentString (node_d, node_d->doc, "deviceID");
      }
   }

#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceResolution::" << __FUNCTION__ << ": returning " << (pszDeviceID_d ? pszDeviceID_d : "NULL") << std::endl;
#endif

   return pszDeviceID_d;
}

bool XMLDeviceResolution::
isResolutionSupported (int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return false;

   xmlDocPtr   docDeviceResolutions  = pXMLDevice->getDocResolutions ();
   xmlNodePtr  rootDeviceResolutions = xmlDocGetRootElement (docDeviceResolutions);
   xmlNodePtr  elmDeviceResolution   = 0;
   bool        fFound                = false;

   if (!rootDeviceResolutions)
      return false;

   elmDeviceResolution = firstNode (rootDeviceResolutions);
   if (!elmDeviceResolution)
      return false;

   elmDeviceResolution = firstNode (elmDeviceResolution->children);

   while (  elmDeviceResolution
         && !fFound
         )
   {
      PSZCRO pszResolution = getXMLContentString (elmDeviceResolution,
                                                  docDeviceResolutions,
                                                  "name");

      if (pszResolution)
      {
         int idResolution = DeviceResolution::nameToID (pszResolution);

         if (idResolution == id)
         {
            fFound = true;
         }

         free ((void *)pszResolution);
      }

      elmDeviceResolution = nextNode (elmDeviceResolution);
   }

   return fFound;
}

class XMLResolution_Enumerator : public Enumeration
{
public:
   XMLResolution_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

      pXMLDevice_d           = pXMLDevice;
      docDeviceResolutions_d = 0;
      nodeItem_d             = nodeItem;

      if (pXMLDevice)
      {
         docDeviceResolutions_d = pXMLDevice->getDocResolutions ();
      }
      else
      {
         nodeItem_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      DeviceResolution *pDRRet = 0;

      if (nodeItem_d)
      {
         PSZCRO pszResolutionName = getXMLContentString (nodeItem_d,
                                                         docDeviceResolutions_d,
                                                         "name");

         if (pszResolutionName)
         {
            pDRRet = XMLDeviceResolution::create (pXMLDevice_d, pszResolutionName);

            free ((void *)pszResolutionName);
         }

         nodeItem_d = nextNode (nodeItem_d);
      }

      return pDRRet;
   }

private:
   XMLDevice  *pXMLDevice_d;
   xmlDocPtr   docDeviceResolutions_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * XMLDeviceResolution::
getEnumeration ()
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return new XMLResolution_Enumerator (pDevice_d, 0);

   xmlDocPtr   docDeviceResolutions  = pXMLDevice->getDocResolutions ();
   xmlNodePtr  rootDeviceResolutions = xmlDocGetRootElement (docDeviceResolutions);
   xmlNodePtr  elmDeviceResolution   = 0;

   if (!rootDeviceResolutions)
      return new XMLResolution_Enumerator (pDevice_d, 0);

   elmDeviceResolution = firstNode (rootDeviceResolutions);
   if (!elmDeviceResolution)
      return new XMLResolution_Enumerator (pDevice_d, 0);

   elmDeviceResolution = firstNode (elmDeviceResolution->children);

   return new XMLResolution_Enumerator (pDevice_d, elmDeviceResolution);
}

std::string XMLDeviceResolution::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DeviceResolution::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const XMLDeviceResolution& const_self)
{
   XMLDeviceResolution& self = const_cast<XMLDeviceResolution&>(const_self);
   std::ostringstream   oss;

   os << self.toString (oss);

   return os;
}
